/*
 * config.c
 *
 * Copyright (C) 1993 Alain Knaff
 */
/*
 *	config.c
 *
 * This file describes how to handle decompression.
 * There are two modes of decompression: pipes and tempfiles
 * Pipes consume less resources than tempfiles, but they don't allow seek and
 * mmap.
 * Fortunately, most programs don't need seek. Thus we can decompress most
 * files through pipes. In order to tell the client that seeks are not allowed,
 * the uncompressed file shows up as a fifo when the client stats it.
 * Some programs might be picky about the file type (fifo or regular) without
 * actually needing seek. That's why an intermediary mode has been provided,
 * which uses a pipe to decompress, but doesn't show it when stat'ing.
 *
 * We have four modes to handle uncompression:
 * PM_USE_TEMP_FILE	decompress into a temporary file
 * PM_HIDE_PIPE		decompress into a pipe, but show the file as regular
 *			when statting it
 * PM_SHOW_PIPE		decompress into a pipe, and show the file as a named
 *			pipe (fifo) when statting it
 * PM_LEAVE_COMPR      	don't decompress at all
 *
 *
 * Files are essentially recognized by their name. An array of FilenameActions
 * structures describes how to open each file. The elements of this array are
 * tried in sequence, and the first matching entry is taken.
 * There are 6 ways to select names:
 *   1. FA_SUFFIX   	matches every file whose name ends in the given suffix
 *   2. FA_BASENAME	matches every file with a given basename. The basename
 *			is the filename without the directory
 *   3. FA_DIR		matches every file which lives in a given directory
 *   4. FA_SUBDIR	matches every file which lives in a given directory or 
 *			in one of its subdirectories.
 *   5. FA_FILENAME	matches every file with a given name, taken in its
 *			entirety
 *   6. FA_ALL		matches every file. This selector must be put into
 *			the last slot of the array.
 *
 * Directory names must be specified with a trailing slash. Patterns are 
 * compared against the names supplied by the client program. These names
 * might not have the form that the user expects! (The user might specify an
 * absolute pathname, but the program tries to access the same file using a
 * relative pathname.)
 */

#include "zlibc.h"

static FilenameActions defaultFilenameActions[]={
  FREC( FA_SUFFIX,	".a",		PM_USE_TMP_FILE), /* static library */
  FREC( FA_SUFFIX,	".sa",		PM_USE_TMP_FILE), /* stub library */
  FREC( FA_SUFFIX,	".o",		PM_USE_TMP_FILE), /* object files */
  FREC( FA_SUFFIX,	".pcf",		PM_USE_TMP_FILE), /* X fonts */
  FREC( FA_SUFFIX,	".pag",		PM_USE_TMP_FILE), /* X rgb.pag */
  FREC( FA_SUFFIX,	".mem",		PM_USE_TMP_FILE),
  FREC( FA_BASENAME,	"CompKeyDB",	PM_USE_TMP_FILE), /* xfig */
  FREC( FA_ALL,		0,		PM_SHOW_PIPE) };

static FilenameActions showPipe[]={
  FREC( FA_ALL,		0,		PM_SHOW_PIPE ) };

static FilenameActions hidePipe[]={
  FREC( FA_ALL,		0,		PM_HIDE_PIPE ) };

static FilenameActions noPipe[]={
  FREC( FA_ALL,		0,		PM_USE_TMP_FILE ) };

static FilenameActions XFilenameActions[]={
  FREC( FA_SUBDIR,     	"/usr/X386",     	PM_USE_TMP_FILE ),
  FREC( FA_SUBDIR,	"/usr/lib/X11",		PM_USE_TMP_FILE ),
  FREC( FA_ALL,		0,			PM_SHOW_PIPE ) };

static FilenameActions emacs[]={
  FREC( FA_SUBDIR,     	"/usr/lib/emacs/",     	PM_SHOW_PIPE ),
  FREC( FA_SUBDIR,	"/usr/info",		PM_SHOW_PIPE ),
  FREC( FA_ALL,		0,			PM_LEAVE_COMPR ) };

/*
 * Certain client programs need a special treatment. These are for instance
 * the programs used to compress and to uncompress files.
 *
 * The first field of this struct contains the basename of the command.
 * The second field of this struct contains a certain number of flags which
 * describes the behaviour of the zlibc when invoked from within these programs.
 *   CM_DISAB		disables the zlibc. 
 *   CM_READDIR_COMPR	readdir shows the compressed (i.e. the real) files
 *			instead of the uncompressed files
 *   CM_VERBOSE		writes comments to stdout while handling compressed 
 *			files. This is essentially useful for debugging.
 *   CM_UNLINK		allows to unlink the real (i.e. compressed file) when
 *			the client unlinks the virtual (uncompressed) file.
 *			As a safety measure, this feature is disabled by 
 *			default.
 *   CM_NONE 		is a place-holder if no other flag is indicated.
 *
 * Several of these flags can be or'ed together. These flags can also be set by
 * environment veriables: ZLIB_DISAB, ZLIB_REEADIR_COMPR, ZLIB_VERBOSE and
 * ZLIB_UNLINK.
 *
 * The third field contains the FilenameAction struct which should be used
 * for these programs.
 *
 * The last entry of this array should contain a null pointer as command name,
 * this matches all commands which have not been matched before.
 */

CommandActions zlib_commandActions[]={

  /* disable the zlibc for compressing programs */
  { "gzip",		CM_DISAB,		0 },
  { "gunzip",		CM_DISAB,		0 },
  { "compress",		CM_DISAB,		0 },
  { "uncompress",	CM_DISAB,		0 },
  { "tar",		CM_UNLINK,		noPipe },

  /* keep rm -r working. */
  { "rm",		CM_UNLINK,	defaultFilenameActions },

  /* keep rm -r working. */
  { "mv",		CM_READDIR_COMPR | CM_UNLINK,	hidePipe },

  /* ls, shows all uncompressed files as pipes, to allow the user to spot
   * the compressed files easily. Don't use this on a Sun, else you lose
   * the size information.
   */
#ifdef linux
  { "ls",		CM_UNLINK,		showPipe },
#endif

  /* find doesn't show pipes. This is useful for constructions like
   * find . -type f -grep pattern {} \;
   * where you don't want to grep through REAL fifos, but where you want to
   * grep through compressed files
   */
  { "find",		CM_UNLINK,		noPipe },

/* less can scroll backwards. Allow it to seek */
  { "less",		CM_UNLINK,		noPipe },

/* file makes several passes (magic and language) Between each pass, it seeks
 * back to the beginning */  
  { "file",		CM_UNLINK,		noPipe },

  { "xdvi",		CM_UNLINK,		noPipe },
  { "xman",       	CM_UNLINK,		noPipe },

  /* lisp can deal with pipes for certain files, but runs much slower.
   * disable pipes altogether for clisp
   */
  { "clisp",       	CM_UNLINK,		noPipe },
  { "lisp.run",		CM_UNLINK,		noPipe },

  /* I don't know too much about smalltalk. Let's be on the safe side :-) */
  { "smalltalk",       	CM_UNLINK,		noPipe },
  { "emacs",		CM_UNLINK,		emacs },

  /* the X-servers don't like pipes either */
  { "XF86_SVGA",	CM_UNLINK,		XFilenameActions },
  { "XF86_VGA16",	CM_UNLINK,		XFilenameActions },
  { "XF86_Mono",	CM_UNLINK,		XFilenameActions },
  { "XF86_S3",		CM_UNLINK,		XFilenameActions },
  { "XF86_Accel",	CM_UNLINK,		XFilenameActions },
  { "X",		CM_UNLINK,		XFilenameActions },

  { 0,			CM_UNLINK,		defaultFilenameActions }
};



char myversion[]="0.4";
